"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.parse = parse;

var _tokenizer = require("./tokenizer");

/**
 * Execution types allow to know what is the query behavior
 *  - LISTING: is when the query list the data
 *  - MODIFICATION: is when the query modificate the database somehow (structure or data)
 *  - INFORMATION: is show some data information such as a profile data
 *  - UNKNOWN
 */
var EXECUTION_TYPES = {
  SELECT: 'LISTING',
  INSERT: 'MODIFICATION',
  DELETE: 'MODIFICATION',
  UPDATE: 'MODIFICATION',
  CREATE_DATABASE: 'MODIFICATION',
  CREATE_TABLE: 'MODIFICATION',
  CREATE_TRIGGER: 'MODIFICATION',
  DROP_DATABASE: 'MODIFICATION',
  DROP_TABLE: 'MODIFICATION',
  TRUNCATE: 'MODIFICATION',
  UNKNOWN: 'UNKNOWN'
};
var dialectsWithEnds = ['sqlite', 'mssql'];
/**
 * Parser
 */

function parse(input) {
  var isStrict = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
  var dialect = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 'generic';
  var topLevelState = initState({
    input: input
  });
  var topLevelStatement = {
    type: 'QUERY',
    start: 0,
    end: input.length - 1,
    body: [],
    tokens: []
  };
  var prevState = topLevelState;
  var statementParser;
  var ignoreOutsideBlankTokens = ['whitespace', 'comment-inline', 'comment-block'];

  while (prevState.position < topLevelState.end) {
    var tokenState = initState({
      prevState: prevState
    });
    var token = (0, _tokenizer.scanToken)(tokenState);

    if (!statementParser) {
      // ignore blank tokens that are not in a statement
      if (ignoreOutsideBlankTokens.includes(token.type)) {
        topLevelStatement.tokens.push(token);
        prevState = tokenState;
        continue;
      }

      statementParser = createStatementParserByToken(token, {
        isStrict: isStrict,
        dialect: dialect
      });
    }

    statementParser.addToken(token);
    topLevelStatement.tokens.push(token);
    prevState = tokenState;
    var statement = statementParser.getStatement();

    if (statement.endStatement) {
      statement.end = token.end;
      topLevelStatement.body.push(statement);
      statementParser = null;
    }
  } // last statement without ending key


  if (statementParser) {
    var _statement = statementParser.getStatement();

    if (!_statement.endStatement) {
      _statement.end = topLevelStatement.end;
      topLevelStatement.body.push(_statement);
    }
  }

  return topLevelStatement;
}

function initState(_ref) {
  var input = _ref.input,
      prevState = _ref.prevState;

  if (prevState) {
    return {
      input: prevState.input,
      position: prevState.position,
      start: prevState.position + 1,
      end: prevState.input.length - 1,
      body: []
    };
  }

  return {
    input: input,
    position: -1,
    start: 0,
    end: input.length - 1,
    body: []
  };
}

function createStatementParserByToken(token, options) {
  if (token.type === 'keyword') {
    switch (token.value.toUpperCase()) {
      case 'SELECT':
        return createSelectStatementParser(options);

      case 'CREATE':
        return createCreateStatementParser(options);

      case 'DROP':
        return createDropStatementParser(options);

      case 'INSERT':
        return createInsertStatementParser(options);

      case 'UPDATE':
        return createUpdateStatementParser(options);

      case 'DELETE':
        return createDeleteStatementParser(options);

      case 'TRUNCATE':
        return createTruncateStatementParser(options);

      default:
        break;
    }
  }

  if (!options.isStrict && token.type === 'unknown') {
    return createUnknownStatementParser(options.isStrict);
  }

  throw new Error("Invalid statement parser \"".concat(token.value, "\""));
}

function createSelectStatementParser(_ref2) {
  var isStrict = _ref2.isStrict;
  var statement = {};
  var steps = [// Select
  {
    preCanGoToNext: function preCanGoToNext() {
      return false;
    },
    validation: {
      acceptTokens: [{
        type: 'keyword',
        value: 'SELECT'
      }]
    },
    add: function add(token) {
      statement.type = 'SELECT';
      statement.start = token.start;
    },
    postCanGoToNext: function postCanGoToNext() {
      return true;
    }
  }];
  return stateMachineStatementParser(statement, steps, {
    isStrict: isStrict
  });
}

function createInsertStatementParser(_ref3) {
  var isStrict = _ref3.isStrict;
  var statement = {};
  var steps = [// Insert
  {
    preCanGoToNext: function preCanGoToNext() {
      return false;
    },
    validation: {
      acceptTokens: [{
        type: 'keyword',
        value: 'INSERT'
      }]
    },
    add: function add(token) {
      statement.type = 'INSERT';
      statement.start = token.start;
    },
    postCanGoToNext: function postCanGoToNext() {
      return true;
    }
  }];
  return stateMachineStatementParser(statement, steps, {
    isStrict: isStrict
  });
}

function createUpdateStatementParser(_ref4) {
  var isStrict = _ref4.isStrict;
  var statement = {};
  var steps = [// Update
  {
    preCanGoToNext: function preCanGoToNext() {
      return false;
    },
    validation: {
      acceptTokens: [{
        type: 'keyword',
        value: 'UPDATE'
      }]
    },
    add: function add(token) {
      statement.type = 'UPDATE';
      statement.start = token.start;
    },
    postCanGoToNext: function postCanGoToNext() {
      return true;
    }
  }];
  return stateMachineStatementParser(statement, steps, {
    isStrict: isStrict
  });
}

function createDeleteStatementParser(_ref5) {
  var isStrict = _ref5.isStrict;
  var statement = {};
  var steps = [// Delete
  {
    preCanGoToNext: function preCanGoToNext() {
      return false;
    },
    validation: {
      acceptTokens: [{
        type: 'keyword',
        value: 'DELETE'
      }]
    },
    add: function add(token) {
      statement.type = 'DELETE';
      statement.start = token.start;
    },
    postCanGoToNext: function postCanGoToNext() {
      return true;
    }
  }];
  return stateMachineStatementParser(statement, steps, {
    isStrict: isStrict
  });
}

function createCreateStatementParser(_ref6) {
  var isStrict = _ref6.isStrict,
      dialect = _ref6.dialect;
  var statement = {};
  var steps = [// Create
  {
    preCanGoToNext: function preCanGoToNext() {
      return false;
    },
    validation: {
      acceptTokens: [{
        type: 'keyword',
        value: 'CREATE'
      }]
    },
    add: function add(token) {
      statement.start = token.start;
    },
    postCanGoToNext: function postCanGoToNext() {
      return true;
    }
  }, // Table/Database
  {
    preCanGoToNext: function preCanGoToNext() {
      return false;
    },
    validation: {
      requireBefore: ['whitespace'],
      acceptTokens: [{
        type: 'keyword',
        value: 'TABLE'
      }, {
        type: 'keyword',
        value: 'DATABASE'
      }, {
        type: 'keyword',
        value: 'TRIGGER'
      }]
    },
    add: function add(token) {
      statement.type = "CREATE_".concat(token.value.toUpperCase());
    },
    postCanGoToNext: function postCanGoToNext() {
      return true;
    }
  }];
  return stateMachineStatementParser(statement, steps, {
    isStrict: isStrict,
    dialect: dialect
  });
}

function createDropStatementParser(_ref7) {
  var isStrict = _ref7.isStrict;
  var statement = {};
  var steps = [// Drop
  {
    preCanGoToNext: function preCanGoToNext() {
      return false;
    },
    validation: {
      acceptTokens: [{
        type: 'keyword',
        value: 'DROP'
      }]
    },
    add: function add(token) {
      statement.start = token.start;
    },
    postCanGoToNext: function postCanGoToNext() {
      return true;
    }
  }, // Table/Database
  {
    preCanGoToNext: function preCanGoToNext() {
      return false;
    },
    validation: {
      requireBefore: ['whitespace'],
      acceptTokens: [{
        type: 'keyword',
        value: 'TABLE'
      }, {
        type: 'keyword',
        value: 'DATABASE'
      }]
    },
    add: function add(token) {
      statement.type = "DROP_".concat(token.value.toUpperCase());
    },
    postCanGoToNext: function postCanGoToNext() {
      return true;
    }
  }];
  return stateMachineStatementParser(statement, steps, {
    isStrict: isStrict
  });
}

function createTruncateStatementParser(_ref8) {
  var isStrict = _ref8.isStrict;
  var statement = {};
  var steps = [{
    preCanGoToNext: function preCanGoToNext() {
      return false;
    },
    validation: {
      acceptTokens: [{
        type: 'keyword',
        value: 'TRUNCATE'
      }]
    },
    add: function add(token) {
      statement.type = 'TRUNCATE';
      statement.start = token.start;
    },
    postCanGoToNext: function postCanGoToNext() {
      return true;
    }
  }];
  return stateMachineStatementParser(statement, steps, {
    isStrict: isStrict
  });
}

function createUnknownStatementParser(_ref9) {
  var isStrict = _ref9.isStrict;
  var statement = {};
  var steps = [{
    preCanGoToNext: function preCanGoToNext() {
      return false;
    },
    add: function add(token) {
      statement.type = 'UNKNOWN';
      statement.start = token.start;
    },
    postCanGoToNext: function postCanGoToNext() {
      return true;
    }
  }];
  return stateMachineStatementParser(statement, steps, {
    isStrict: isStrict
  });
}

function stateMachineStatementParser(statement, steps, _ref10) {
  var isStrict = _ref10.isStrict,
      _ref10$dialect = _ref10.dialect,
      dialect = _ref10$dialect === void 0 ? 'generic' : _ref10$dialect;
  var currentStepIndex = 0;
  var prevToken;
  /* eslint arrow-body-style: 0, no-extra-parens: 0 */

  var isValidToken = function isValidToken(step, token) {
    if (!step.validation) {
      return true;
    }

    return step.validation.acceptTokens.filter(function (accept) {
      var isValidType = token.type === accept.type;
      var isValidValue = !accept.value || token.value.toUpperCase() === accept.value;
      return isValidType && isValidValue;
    }).length > 0;
  };

  var hasRequiredBefore = function hasRequiredBefore(step) {
    return !step.requireBefore || step.requireBefore.includes(prevToken.type);
  };

  return {
    getStatement: function getStatement() {
      return statement;
    },
    addToken: function addToken(token) {
      /* eslint no-param-reassign: 0 */
      if (statement.endStatement) {
        throw new Error('This statement has already got to the end.');
      }

      if (token.type === 'semicolon') {
        // SQLite and MSSQL require semi-colons inside the trigger. They signify the end of the trigger creation
        // with `END;`. This allows detection of that.
        if (dialectsWithEnds.includes(dialect) && statement.type === 'CREATE_TRIGGER' && !statement.canEnd) {// do nothing
        } else {
          statement.endStatement = ';';
          return;
        }
      } // SQLite and MSSQL triggers use `END;` to signify the end of the statement. The statement can include other semicolons.


      if (dialectsWithEnds.includes(dialect) && token.value === 'END' && statement.type === 'CREATE_TRIGGER') {
        statement.canEnd = true;
        return;
      }

      if (token.type === 'whitespace') {
        prevToken = token;
        return;
      }

      if (statement.type) {
        // statement has already been identified
        // just wait until end of the statement
        return;
      }

      var currentStep = steps[currentStepIndex];

      if (currentStep.preCanGoToNext(token)) {
        currentStepIndex++;
        currentStep = steps[currentStepIndex];
      }

      if (!hasRequiredBefore(currentStep)) {
        var requireds = currentStep.requireBefore.join(' or ');
        throw new Error("Expected any of these tokens ".concat(requireds, " before \"").concat(token.value, "\" (currentStep=").concat(currentStepIndex, ")."));
      }

      if (!isValidToken(currentStep, token) && isStrict) {
        var expecteds = currentStep.validation.acceptTokens.map(function (accept) {
          return "(type=\"".concat(accept.type, "\" value=\"").concat(accept.value, "\")");
        }).join(' or ');
        throw new Error("Expected any of these tokens ".concat(expecteds, " instead of type=\"").concat(token.type, "\" value=\"").concat(token.value, "\" (currentStep=").concat(currentStepIndex, ")."));
      }

      currentStep.add(token);
      statement.executionType = EXECUTION_TYPES[statement.type] || 'UNKNOWN';

      if (currentStep.postCanGoToNext(token)) {
        currentStepIndex++;
      }

      prevToken = token;
    }
  };
}